/*  Screem:  pageUI.c,
 * 
 *  ScreemPage utility functions for gui stuff
 * 
 *  Copyright (C) 1999 - 2002  David A Knight
 *
 *  screem_print_page() (and all the other printing code now)
 *  is derived from the code in gedit-print.c of
 *  Gedit 2, which is Copyright (C) 2000, 2001 Chema Celorio, Paolo Maggi
 *  and the gedit Team 1998-2001. See the Gedit AUTHORS file for a list
 *  of people on the gedit Team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-url.h>
#include <libgnomeui/gnome-messagebox.h>
#include <libgnomeui/gnome-stock-icons.h>
#include <libgnomeui/gnome-entry.h>

#include <libgnomevfs/gnome-vfs-uri.h>
#include <libgnomevfs/gnome-vfs-ops.h>
#include <libgnomevfs/gnome-vfs-file-info.h>
#include <libgnomevfs/gnome-vfs-utils.h>

#include <bonobo/bonobo-file-selector-util.h>

#include <gconf/gconf-client.h>

#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include <glade/glade.h>

#include <gtk/gtkinvisible.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtktreemodel.h>
#include <gtk/gtkcellrenderer.h>
#include <gtk/gtkcellrenderertext.h>

#include "editMenu.h"
#include "fileops.h"
#include "pageUI.h"
#include "support.h"

#include "screem-window.h"
#include "screem-site.h"
#include "screem-site-ui.h"
#include "screem-editor.h"
#include "screem-preview.h"
#include "screem-page.h"

#include "screem-print.h"

static gboolean screem_print_run_dialog( ScreemPrintJob *job );
static void screem_page_print_header( ScreemPrintJob *pji, gint page_number );
static void screem_print_document( ScreemPrintJob *pji );
static void screem_print_preview_real( ScreemPrintJob *pji );
static void screem_print_real( ScreemPage* page, gboolean preview, 
			       GError **error );
static void screem_print_error_dialog( GError *error );
static const gchar*
screem_print_get_next_line_to_print_delimiter( ScreemPrintJob *pji,
					       const gchar *start, 
					       const gchar *end, 
					       gboolean first_line_of_par );
static void screem_print_line_number( ScreemPrintJob *pji, gdouble y );
static void screem_print_line( ScreemPrintJob *pji, 
			       const gchar *start, const gchar *end, gdouble y,
			       gboolean first_line_of_par );
static gdouble screem_print_paragraph( ScreemPrintJob *pji, 
				       const gchar *start, const gchar *end, 
				       gdouble y );
static gdouble screem_print_create_new_page( ScreemPrintJob *pji );
static void screem_print_end_page( ScreemPrintJob *pji );
static void screem_print_update_page_size_and_margins( ScreemPrintJob *pji );


gboolean screem_page_open_with_filename( ScreemSite *site, 
					 ScreemWindow *window,
					 const gchar *filename )
{
	ScreemApplication *app;
	ScreemPage *page;
	gchar *path;
	GnomeVFSURI *uri;

	g_return_val_if_fail( filename != NULL, FALSE );

	g_object_get( G_OBJECT( site ), "app", &app, NULL );
	
	uri = gnome_vfs_uri_new( filename );

	if( ! uri ) {
		return FALSE;
	}
	
	/* is the filename local? */
	if( gnome_vfs_uri_is_local( uri ) ) {
		/* yes */
		/* HACK */
		if( *filename == G_DIR_SEPARATOR ) {
			path = g_strconcat( "file://", filename, NULL );
		} else {
			path = relative_to_full( filename, NULL );
		}
	} else {
		path = g_strdup( filename );
	}
	gnome_vfs_uri_unref( uri );
	

	page = screem_site_locate_page( site, path );

	if( screem_site_get_fake_flag( site ) && ! page ) {
		/* fake site, add to list of pages */
		page = screem_page_new( G_OBJECT( app ) );
		screem_page_set_pathname( page, path );
		
		if( ! screem_site_add_page( site, page ) ) {
			/* page is already loaded, FIXME: what now? */
		}
	}

	g_free( path );

	if( page ) {
		if( screem_window_get_current( window ) != site ) {
			screem_window_set_current( window, site );
		}
		screem_window_set_document( window, page );
	}

	return ( page != NULL );
}


gboolean screem_page_save_as( ScreemPage *page, ScreemWindow *window )
{
	gchar *filename;
	gboolean hold;
	ScreemSite *site;
	const gchar *pathname;
	gchar *spath;

	site = screem_window_get_current( window );
	pathname = screem_site_get_pathname( site );
	spath = NULL;
	if( ! screem_site_get_fake_flag( site ) ) {
		GnomeVFSURI *uri;
		pathname = screem_site_get_pathname( site );
		uri = gnome_vfs_uri_new( pathname );
		if( ! uri ) {
			spath = NULL;
		} else if( gnome_vfs_uri_is_local( uri ) ) {
			spath = gnome_vfs_get_local_path_from_uri( pathname );
			gnome_vfs_uri_unref( uri );
		} else {
			spath = gnome_vfs_uri_to_string( uri, 0 );
			gnome_vfs_uri_unref( uri );
		}
	}

	filename = 
		bonobo_file_selector_save( GTK_WINDOW( window ), TRUE,
					   _( "Select filename to save as" ),
					   NULL, spath, NULL );

	/* we didn't select one */
	hold = FALSE;
	if( filename ) {
		if( ! overwrite( filename ) ) {
			g_free( filename );
			hold = FALSE;
		} else {
			screem_page_set_pathname( page, filename );
			
			/* allow save by setting page as changed */
			screem_page_set_changed( page, TRUE );
			hold = screem_page_save( page );
		}
	}

	if( spath ) {
		g_free( spath );
	}
	
	return hold;
}

gint screem_page_save_confirmation( ScreemPage *page )
{
	GtkWidget *box;
        gint button;
	const gchar *pathname;
	gchar *temp;
	
	if( ! screem_page_get_data( page ) ) {
		return GTK_RESPONSE_NO;
	} else if( ! screem_page_get_changed( page ) ) {
		return GTK_RESPONSE_NO;
	}

	pathname = screem_page_get_pathname( page );

	if( ! pathname ) {
		pathname = _( "Untitled" );
	}
	temp = gnome_vfs_unescape_string_for_display( pathname );

	box = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL,
				      GTK_MESSAGE_QUESTION,
				      GTK_BUTTONS_YES_NO,
				      _( "Save: %s?" ),
				      temp );
	g_free( temp );
	gtk_window_set_wmclass( GTK_WINDOW( box ), "Screem",
				"save_confirm_prompt" );
        button = gtk_dialog_run( GTK_DIALOG( box ) );

	gtk_widget_destroy( box );
					      
        if( button == GTK_RESPONSE_YES ) {
                screem_page_save( page );
	} else if( button == GTK_RESPONSE_NONE ) {
		button = GTK_RESPONSE_CANCEL;
	}

	return button;
}

void screem_page_close_page( ScreemPage *page, ScreemWindow *window )
{
	/* request save confirmation */
	if( screem_page_save_confirmation( page ) != GTK_RESPONSE_CANCEL ) {
		screem_page_set_open( page, FALSE );
		screem_window_remove_document( window, page );
	}
}

gboolean screem_page_save_check( ScreemPage *page )
{
	const gchar *path;
	gint button = GTK_RESPONSE_NO;
	GtkWidget *prompt;
	GnomeVFSFileInfo *info;
	GnomeVFSURI *uri;
	GnomeVFSResult res;

	path = screem_page_get_pathname( page );

	/* Don't save if there is no file associated with this page yet */
	if( ! path ) {
		return FALSE;
	}

	uri = gnome_vfs_uri_new( path );

	if( uri ) {
		info = gnome_vfs_file_info_new();
		res = gnome_vfs_get_file_info_uri( uri, info,
						   GNOME_VFS_FILE_INFO_FOLLOW_LINKS );
		gnome_vfs_uri_unref( uri );
	} else {
		info = NULL;
		res = GNOME_VFS_ERROR_GENERIC;
	}

	if( res != GNOME_VFS_OK && res != GNOME_VFS_ERROR_NOT_FOUND ) {
		if( info ) {
			gnome_vfs_file_info_unref( info );
		}
		return FALSE;
	}

	/* file exists, check if it has been modified outside of screem */
	if( screem_page_get_modified( page ) ) {
		gchar *temp;
		temp = gnome_vfs_unescape_string_for_display( path );
		prompt = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL,
						 GTK_MESSAGE_WARNING,
						 GTK_BUTTONS_YES_NO,
						 _("%s has changed on disk\nReload instead of saving?"), temp );
		g_free( temp );
		gtk_window_set_wmclass( GTK_WINDOW( prompt ), "Screem",
					"revert_prompt" );

                button = gtk_dialog_run( GTK_DIALOG( prompt ) );

		gtk_widget_destroy( prompt );

		if( button == GTK_RESPONSE_YES ) {
			screem_page_revert( page );
		}
	}

	gnome_vfs_file_info_unref( info );

	return (button == GTK_RESPONSE_NO);
}

void external_view( ScreemPage *page )
{
        const gchar *pathname;
	gchar *uri;

        pathname = screem_page_get_pathname( page );

        g_return_if_fail( pathname != NULL );

        if( screem_page_save_confirmation( page ) != GTK_RESPONSE_CANCEL ) {
		uri = g_strconcat( "file://", pathname, NULL );
		
		gnome_url_show( uri, NULL );
		
		g_free( uri );
	}
}


void screem_page_edit_bookmarks( ScreemWindow *window, ScreemEditor *editor,
				 ScreemPage *page )
{
	GladeXML *xml;
	GConfClient *client;
	gchar *gladepath;

	GtkWidget *dialog;
	gint response;
	GtkWidget *bookmarkname;
	GtkWidget *bookmarkline;
	GtkWidget *jumpto;
	const gchar *name;
	gchar *title;
	gint line;
	gboolean allowjump;
	const gchar *pathname;
	gchar *temp;

	client = gconf_client_get_default();
	gladepath = gconf_client_get_string( client, "/apps/screem/general/glade_path",
					     NULL );
	xml = glade_xml_new( gladepath, "bookmarks", NULL );
	g_free( gladepath );

	dialog = glade_xml_get_widget( xml, "bookmarks" );
	bookmarkname = glade_xml_get_widget( xml, "bookmarkname" );
	bookmarkname = gnome_entry_gtk_entry( GNOME_ENTRY( bookmarkname ) );

	bookmarkline = glade_xml_get_widget( xml, "bookmarkline" );
	line = screem_page_get_line_count( page );
	gtk_spin_button_set_range( GTK_SPIN_BUTTON( bookmarkline ),
				   1.0, (gdouble)line );

	jumpto = glade_xml_get_widget( xml, "jumptobookmark" );
	allowjump = ( page == screem_window_get_document( window ) );
	gtk_widget_set_sensitive( jumpto, allowjump );

	{
		GtkWidget *treeview;
		GtkCellRenderer *renderer;
		GtkTreeViewColumn *col;
		GtkListStore *store;

		treeview = glade_xml_get_widget( xml, "bookmarkview" );
		g_object_set_data( G_OBJECT( treeview ), "page", page );
		
		renderer = gtk_cell_renderer_text_new();
		col = gtk_tree_view_column_new();
		gtk_tree_view_column_set_title( col, "Name" );
		gtk_tree_view_column_pack_start( col, renderer, TRUE );
		gtk_tree_view_append_column( GTK_TREE_VIEW( treeview ),
					     col );
		gtk_tree_view_column_set_attributes( col, renderer,
						     "text", 0, NULL );

		renderer = gtk_cell_renderer_text_new();
		col = gtk_tree_view_column_new();
		gtk_tree_view_column_set_title( col, "Line" );
		gtk_tree_view_column_pack_start( col, renderer, TRUE );
		gtk_tree_view_append_column( GTK_TREE_VIEW( treeview ),
					     col );
		gtk_tree_view_column_set_attributes( col, renderer,
						     "text", 1, NULL );

		store = screem_page_get_bookmark_model( page );
		gtk_tree_view_set_model( GTK_TREE_VIEW( treeview ),
					 GTK_TREE_MODEL( store ) );
	}


	pathname = screem_page_get_pathname( page );
	if( ! pathname ) {
		pathname = _( "Untitled" );
	}
	temp = gnome_vfs_unescape_string_for_display( pathname );

	title = g_strconcat( temp, "- Screem", NULL );
	gtk_window_set_title( GTK_WINDOW( dialog ), title );
	g_free( temp );
	g_free( title );

	glade_xml_signal_autoconnect( xml );

	while( dialog ) {
		response = gtk_dialog_run( GTK_DIALOG( dialog ) );
		
		name = gtk_entry_get_text( GTK_ENTRY( bookmarkname ) );
		line = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(bookmarkline));

		switch( response ) {
		case 0:
			/* add */
			screem_page_add_bookmark( page, line, name );
			screem_editor_expose_lines( editor );
			break;
		case 1:
			/* remove */
			screem_page_remove_bookmark( page, name );
			screem_editor_expose_lines( editor );
			break;
		case 2:
			/* jump to */
			line = screem_page_get_bookmark( page, name );
			if( line != -1 )
				screem_editor_goto_line( editor, line );
			break;
		default:
			break;
		}
		
		if( response != 0 && response != 1 ) {
			gtk_widget_destroy( dialog );
			dialog = NULL;
		}
	}
}

void screem_page_bookmark_activate( GtkTreeView *view,
				    GtkTreePath *path,
				    GtkTreeViewColumn *column,
				    gpointer data )
{
	GtkTreeSelection *selection;
	GtkListStore *model;
	GtkTreeModel *tmodel;
	GladeXML *xml;
	ScreemPage *page;

	xml = glade_get_widget_tree( GTK_WIDGET( view ) );
	selection = gtk_tree_view_get_selection( view );
	if( selection ) {
		GtkTreeIter it;
		GValue value = {0,};
		ScreemPageBookMark *bookmark;
		GtkWidget *name;
		GtkWidget *line;

		page = SCREEM_PAGE( g_object_get_data( G_OBJECT( view ), 
						       "page" ) );
		model = screem_page_get_bookmark_model( page );
		tmodel = GTK_TREE_MODEL( model );

		gtk_tree_selection_get_selected( selection, &tmodel, &it );
		gtk_tree_model_get_value( GTK_TREE_MODEL( model ),
					  &it, 2, &value );
		bookmark = g_value_get_pointer( &value );

		name = glade_xml_get_widget( xml, "bookmarkname" );
		line = glade_xml_get_widget( xml, "bookmarkline" );

		name = gnome_entry_gtk_entry( GNOME_ENTRY( name ) );
		gtk_entry_set_text( GTK_ENTRY( name ), 
				    bookmark->name->str );

		gtk_spin_button_set_value( GTK_SPIN_BUTTON( line ),
					   atof( bookmark->line_number ) );
	}
}

void screem_page_print( ScreemPage *page, gboolean preview )
{
	GError *error = NULL;
	
        g_return_if_fail( page != NULL );

        screem_print_real( page, preview, &error );
	
        if( error ) {
                screem_print_error_dialog( error );
                g_error_free( error );
        }
}

static gboolean screem_print_run_dialog( ScreemPrintJob *job )
{
	GtkWidget *dialog;
	gint lines;
	gint res;

        g_return_val_if_fail( job != NULL, TRUE );

	dialog = g_object_new( GNOME_TYPE_PRINT_DIALOG, 
			       "print_config", job->config, NULL );
	
	gnome_print_dialog_construct( GNOME_PRINT_DIALOG( dialog ),
				      _( "Screem - Print Document " ),
				      GNOME_PRINT_DIALOG_RANGE |
				      GNOME_PRINT_DIALOG_COPIES );

	lines = screem_page_get_line_count( job->page );

	gnome_print_dialog_construct_range_page( GNOME_PRINT_DIALOG( dialog ),
						 GNOME_PRINT_RANGE_ALL,// |
						 //GNOME_PRINT_RANGE_RANGE
						 1, lines, "A", "Lines" );

	res = gtk_dialog_run( GTK_DIALOG( dialog ) );
	switch( res ) {
	case GNOME_PRINT_DIALOG_RESPONSE_PRINT:
		job->preview = FALSE;
		break;
	case GNOME_PRINT_DIALOG_RESPONSE_PREVIEW:
		job->preview = TRUE;
		break;
	default:
		gtk_widget_destroy( dialog );
		return TRUE;
		break;
	}

	job->range_type = gnome_print_dialog_get_range( GNOME_PRINT_DIALOG( dialog ) );

	gtk_widget_destroy( dialog );

	return FALSE;
}

static void screem_page_print_header( ScreemPrintJob *pji, gint page_number )
{
        gdouble y, x, len;
        gchar* l_text;
        gchar* r_text;
	const gchar *pathname;
	gchar *temp;

        g_return_if_fail( pji != NULL );
        g_return_if_fail( pji->font_header != NULL );
        g_return_if_fail( pji->page != NULL );

        gnome_print_setfont( pji->print_ctx, pji->font_header );

        y = pji->page_height - pji->margin_top -
                gnome_font_get_ascender( pji->font_header );

        /* Print left text */
        x = pji->margin_right;

	pathname = screem_page_get_pathname( pji->page );
	if( ! pathname ) {
		pathname = _( "Untitled" );
	}
	temp = gnome_vfs_unescape_string_for_display( pathname );

        l_text = g_strconcat( "File: ", temp, NULL );
	g_free( temp );
        gnome_print_moveto( pji->print_ctx, x, y );
        gnome_print_show( pji->print_ctx, l_text );

        /* Print right text */
        r_text = g_strdup_printf( _( "Page: %d" ), page_number );
        len = gnome_font_get_width_utf8( pji->font_header, r_text );

        x = pji->page_width - pji->margin_right - len;

        gnome_print_moveto( pji->print_ctx, x, y );
        gnome_print_show( pji->print_ctx, r_text );

        /* Print line */
        y = pji->page_height - pji->margin_top -
                ( 1.5 * gnome_font_get_size( pji->font_header ) );

        gnome_print_setlinewidth( pji->print_ctx, 1.0 );
        gnome_print_moveto( pji->print_ctx,  pji->margin_right, y );
        gnome_print_lineto( pji->print_ctx, 
			    pji->page_width - pji->margin_right, y );
        gnome_print_stroke( pji->print_ctx );

        g_free( l_text );
        g_free( r_text );

        gnome_print_setfont( pji->print_ctx, pji->font_body );
}

static void screem_print_document( ScreemPrintJob *pji )
{
        gchar *text_to_print;
        gchar *text_end;
        const gchar *end;
        const gchar *p;
        gdouble y;
        gint paragraph_delimiter_index;
        gint next_paragraph_start;
        gdouble fontheight;

        switch( pji->range_type ) {
	case GNOME_PRINT_RANGE_ALL:
		text_to_print = screem_page_get_data( pji->page );
		text_end = text_to_print + strlen( text_to_print );
		break;
	case GNOME_PRINT_RANGE_SELECTION:
		text_to_print = screem_page_get_data( pji->page );
		text_end = text_to_print + strlen( text_to_print );

		pji->first_line_to_print = 1;
		/*screem_page_get_line_at_offset( pji->page, 
		  0 ) + 1;*/
		break;
	default:
		g_return_if_fail( FALSE );
        }

        y = screem_print_create_new_page( pji );

        gnome_print_setfont( pji->print_ctx, pji->font_body );

        pango_find_paragraph_boundary( text_to_print, -1,
				       &paragraph_delimiter_index,
				       &next_paragraph_start );

        p = text_to_print;
        end = text_to_print + paragraph_delimiter_index;

        fontheight = gnome_font_get_ascender( pji->font_body ) +
		gnome_font_get_descender( pji->font_body );

        while( p < text_end ) {
                gunichar wc = g_utf8_get_char( p );
                /* Only one character has type G_UNICODE_PARAGRAPH_SEPARATOR in
                 * Unicode 3.0; update this if that changes.
                 */
#define PARAGRAPH_SEPARATOR 0x2029

                if( ( wc == '\n' || wc == '\r' || wc == PARAGRAPH_SEPARATOR )){
			if( pji->numbers_width > 0.0 )
                                screem_print_line_number( pji, y );

                        y -= 1.2 * gnome_font_get_size( pji->font_body );

                        if( ( y - pji->margin_bottom ) < fontheight ) {
                                screem_print_end_page( pji );

                                y = screem_print_create_new_page( pji );
                        }
                } else {
                        y = screem_print_paragraph( pji, p, end, y );
                }

                ++ pji->printed_lines;

                p = p + next_paragraph_start;

                if( p < text_end ) {
                        pango_find_paragraph_boundary( p, -1,
						       &paragraph_delimiter_index, &next_paragraph_start );
                        end = p + paragraph_delimiter_index;
                }

        }

        gnome_print_showpage( pji->print_ctx );

        g_free( text_to_print );
}

/**
 * screem_print_preview_real:
 * @pji: 
 * 
 * Create and show the print preview window
 **/
static void screem_print_preview_real( ScreemPrintJob *pji )
{
        GtkWidget *gpmp;
        gchar *title;
	const gchar *pathname;
	gchar *temp;

        g_return_if_fail( pji != NULL );
        g_return_if_fail( pji->print_job != NULL );

	pathname = screem_page_get_pathname( pji->page );
	if( ! pathname ) {
		pathname = _( "Untitled" );
	}
	temp = gnome_vfs_unescape_string_for_display( pathname );

        title = g_strconcat( temp, "- Screem Print Preview", NULL );
        gpmp = gnome_print_job_preview_new( pji->print_job, title );
        g_free( title );

        gtk_widget_show( gpmp );
}

/**
 * screem_print_real:
 * @preview: 
 * 
 * The main printing function
 **/
static void screem_print_real( ScreemPage* page, gboolean preview, 
			       GError **error )
{
        ScreemPrintJob *pji;
        gboolean cancel = FALSE;

	pji = screem_print_job_new( page, preview );

        if( pji == NULL )
                return;

        if( ! pji->preview ) {
                /* 
                 * Open the print dialog and initialize
                 * pji->config 
                 */
                cancel = screem_print_run_dialog( pji );
	}

        /* The canceled button on the dialog was clicked */
        if( cancel ) {
                screem_print_job_destroy( pji );
                return;
        }

        g_return_if_fail( pji->config != NULL );

        pji->print_job = gnome_print_job_new( pji->config );
        g_return_if_fail( pji->print_job != NULL );

        pji->print_ctx = gnome_print_job_get_context( pji->print_job );
        g_return_if_fail( pji->print_ctx != NULL );

        screem_print_update_page_size_and_margins( pji );
        screem_print_document( pji );
#if 0
        /* The printing was canceled while in progress */
        if( pji->canceled ) {
                screem_print_job_destroy( pji );
                return;
        }
#endif  
        gnome_print_job_close( pji->print_job );

        if( pji->preview ) 
                screem_print_preview_real( pji );
        else
                gnome_print_job_print( pji->print_job );

        screem_print_job_destroy( pji );
}

static void screem_print_error_dialog( GError *error )
{
        GtkWidget *dialog;
        gchar *errstr;
                        
        errstr = g_strdup_printf(_("An error occurred while printing.\n\n%s,"),
				 error->message );
                 
        dialog = gtk_message_dialog_new( NULL,
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_OK,
					 errstr );
	
        gtk_dialog_set_default_response( GTK_DIALOG( dialog ), 
					 GTK_RESPONSE_OK );

        gtk_window_set_resizable( GTK_WINDOW( dialog ), FALSE );

        gtk_dialog_run( GTK_DIALOG( dialog ) );

        gtk_widget_destroy( dialog );
                        
        g_free( errstr );
}


static const gchar*
screem_print_get_next_line_to_print_delimiter( ScreemPrintJob *pji,
					       const gchar *start, 
					       const gchar *end, 
					       gboolean first_line_of_par )
{
        const gchar* p;
        gdouble line_width = 0.0;
        gdouble printable_page_width;
        ArtPoint space_advance;
        gint space;

        gint chars_in_this_line = 0;
        gint tab_size = 8;  /* FIXME: should be configurable */

        if( pji->numbers_width > 0.0 ) {
                printable_page_width = pji->page_width -
                        ( pji->margin_right + pji->margin_left ) - 
			pji->numbers_width - CM( 0.5 );
        } else {
                printable_page_width = pji->page_width -
                        ( pji->margin_right + pji->margin_left );
	}

        p = start;

        /* Find space advance */
        space = gnome_font_lookup_default( pji->font_body, ' ' );
        gnome_font_get_glyph_stdadvance( pji->font_body, space, 
					 &space_advance );
        if( ! first_line_of_par ) {
		gunichar ch = g_utf8_get_char( p );
                
                while( ch == ' ' || ch == '\t' ) {
                        p = g_utf8_next_char( p );
                        ch = g_utf8_get_char( p );
                }
        }

        while( p < end ) {
		gunichar ch;
                gint glyph;
                
                ch = g_utf8_get_char( p );
                ++chars_in_this_line;
                
                if( ch == '\t' ) {
                        /* FIXME: use a tabs array */
                        gint num_of_equivalent_spaces;

                        num_of_equivalent_spaces = tab_size - 
				( ( chars_in_this_line - 1 ) % tab_size );
                        chars_in_this_line += num_of_equivalent_spaces - 1;
                        
                        line_width += ( num_of_equivalent_spaces * space_advance.x );
                } else {
                        glyph = gnome_font_lookup_default( pji->font_body, 
							   ch );
                        /* FIXME */
                        if( glyph == space ) {
                                line_width += space_advance.x;
                        } else {
				/*
				  if ((glyph < 0) || (glyph >= 256))
				*/
				ArtPoint adv;
                                gnome_font_get_glyph_stdadvance(pji->font_body,
								glyph, &adv );
                                if( adv.x > 0 ) {
                                        line_width += adv.x;
				} else {
                                        /* To be as conservative as possible */
                                        line_width += ( 2 * space_advance.x );
				}
                        }  /*else {
			     line_width += gnome_font_get_glyph_width (pji->font_body, glyph);
			     }
			   */
                }

                if( line_width > printable_page_width ) {
                        /* FIXME: take care of lines wrapping at word 
			   boundaries too */
                        return p;
                }
                p = g_utf8_next_char( p );
        }
	
        return end;
}

static void screem_print_line_number( ScreemPrintJob *pji, gdouble y )
{
        gchar *num_str;
        gdouble len;
        gdouble x;

        gint line_num = pji->first_line_to_print + pji->printed_lines;

        if( ( pji->printed_lines + 1 ) % pji->print_line_numbers != 0 ) {
		return;
	}

        num_str = g_strdup_printf( "%d", line_num );

        gnome_print_setfont( pji->print_ctx, pji->font_numbers );

        len = gnome_font_get_width_utf8( pji->font_numbers, num_str );

        x = pji->margin_left + pji->numbers_width - len;

        gnome_print_moveto( pji->print_ctx, x, y -
			    gnome_font_get_ascender( pji->font_numbers ) );
        gnome_print_show( pji->print_ctx, num_str );

        g_free( num_str );

        /* Restore body font */
        gnome_print_setfont( pji->print_ctx, pji->font_body );
}

static void screem_print_line( ScreemPrintJob *pji, 
			       const gchar *start, const gchar *end, gdouble y,
			       gboolean first_line_of_par )
{
        GnomeGlyphList * gl;
        gdouble x;
        const gchar* p;
        gint chars_in_this_line = 0;
        gint tab_size = 8; /* FIXME should be configurable*/

        p = start;

        gl = gnome_glyphlist_from_text_dumb( pji->font_body, 
					     0x000000ff, 0.0, 0.0, "" );

        gnome_glyphlist_advance( gl, TRUE );

        if( pji->numbers_width > 0.0 ) {
                if( first_line_of_par ) {
                        screem_print_line_number( pji, y );
		}

                x = pji->margin_left + pji->numbers_width + CM( 0.5 );
        } else {
                x = pji->margin_left;
	}

        gnome_glyphlist_moveto( gl, x, 
				y - 
				gnome_font_get_ascender( pji->font_body ) );

        if( ! first_line_of_par ) {
                gunichar ch = g_utf8_get_char( p );

                while( ch == ' ' || ch == '\t' ) {
                        p = g_utf8_next_char( p );
                        ch = g_utf8_get_char( p );
                }
        }

        while( p < end ) {
		gunichar ch;
                gint glyph;

                ch = g_utf8_get_char( p );
                ++ chars_in_this_line;

                if( ch == '\t' ) {
                        /* FIXME: use a tabs array */
                        gint i;
                        gint num_of_equivalent_spaces;
			
                        glyph = gnome_font_lookup_default( pji->font_body, 
							   ' ' );
                        num_of_equivalent_spaces = tab_size - 
				( ( chars_in_this_line - 1 ) % tab_size );
                        chars_in_this_line += num_of_equivalent_spaces - 1;

                        for( i = 0; i < num_of_equivalent_spaces; ++i ) {
                                gnome_glyphlist_glyph( gl, glyph );
			}
                } else {
                        glyph = gnome_font_lookup_default( pji->font_body, 
							   ch );
                        gnome_glyphlist_glyph( gl, glyph );
                }

                p = g_utf8_next_char( p );
        }

        gnome_print_moveto( pji->print_ctx, 0.0, 0.0 );
        gnome_print_glyphlist( pji->print_ctx, gl );
        gnome_glyphlist_unref( gl );
}


static gdouble screem_print_paragraph( ScreemPrintJob *pji, 
				       const gchar *start, const gchar *end,
				       gdouble y )
{
        const gchar *p, *s;
        gdouble fontheight;
        gboolean fl;

        p = start;

        fl = TRUE;

        fontheight = gnome_font_get_ascender( pji->font_body ) +
		gnome_font_get_descender( pji->font_body );

        while( p < end ) {
                s = p;
                p = screem_print_get_next_line_to_print_delimiter( pji, 
								   s, end, 
								   fl );
                screem_print_line( pji, s, p, y, fl );
                y -= 1.2 * gnome_font_get_size( pji->font_body );
                fl = FALSE;

                if( ( y - pji->margin_bottom ) < fontheight ) {
                        screem_print_end_page( pji );

                        y = screem_print_create_new_page( pji );
                }

                if( pji->print_wrap_lines ) {
                        return y;
		}
        }

        return y;
}

static gdouble screem_print_create_new_page( ScreemPrintJob *pji )
{
        gchar *pn_str;

        ++ pji->page_num;
        pn_str = g_strdup_printf( "%d", pji->page_num );
        gnome_print_beginpage( pji->print_ctx, pn_str );

        g_free( pn_str );

        if( pji->print_header ) {
                screem_page_print_header( pji, pji->page_num );
                return pji->page_height - pji->margin_top - pji->header_height;
        } else {
                return pji->page_height - pji->margin_top;
	}
}

static void screem_print_end_page( ScreemPrintJob *pji )
{
        gnome_print_showpage( pji->print_ctx );
}

static void screem_print_update_page_size_and_margins( ScreemPrintJob *pji )
{
        const GnomePrintUnit *unit;
                
	gnome_print_job_get_page_size_from_config( pji->config, 
						   &pji->page_width, 
						   &pji->page_height );
	
        if( gnome_print_config_get_length( pji->config, 
					   GNOME_PRINT_KEY_PAGE_MARGIN_LEFT, 
					   &pji->margin_left, &unit ) ) {
                gnome_print_convert_distance( &pji->margin_left, 
					      unit, GNOME_PRINT_PS_UNIT );
        }
        
        if( gnome_print_config_get_length( pji->config, 
					   GNOME_PRINT_KEY_PAGE_MARGIN_RIGHT, 
					   &pji->margin_right, &unit ) ) {
		gnome_print_convert_distance( &pji->margin_right, 
					      unit, GNOME_PRINT_PS_UNIT );
        }
        
        if( gnome_print_config_get_length( pji->config, 
					   GNOME_PRINT_KEY_PAGE_MARGIN_TOP, 
					   &pji->margin_top, &unit ) ) {
		gnome_print_convert_distance( &pji->margin_top, 
					      unit, GNOME_PRINT_PS_UNIT );
        }
	
        if( gnome_print_config_get_length( pji->config, 
					   GNOME_PRINT_KEY_PAGE_MARGIN_BOTTOM, 
					   &pji->margin_bottom, &unit ) ) {
		gnome_print_convert_distance( &pji->margin_bottom, 
					      unit, GNOME_PRINT_PS_UNIT );
        }

        if( pji->print_line_numbers ) {
                gchar* num_str;

                num_str = 
			g_strdup_printf("%d", 
					screem_page_get_line_count(pji->page));
                pji->numbers_width = gnome_font_get_width_utf8( pji->font_numbers, num_str );
                g_free( num_str );
        }
}
